import math as m
import yaml
import logging

from typing import List, Tuple
from OSMPythonTools.overpass import Overpass, overpassQueryBuilder
from OSMPythonTools import cachingStrategy
from pywikibot import ItemPage, Site
from pywikibot import config
config.put_throttle = 0
config.maxlag = 0

from structs.preferences import Preferences, Preference
from structs.landmarks import Landmark
from utils import take_most_important
import constants


SIGHTSEEING = 'sightseeing'
NATURE = 'nature'
SHOPPING = 'shopping'



class LandmarkManager:

    logger = logging.getLogger(__name__)

    city_bbox_side: int     # bbox side in meters
    radius_close_to: int    # radius in meters
    church_coeff: float     # coeff to adjsut score of churches
    park_coeff: float       # coeff to adjust score of parks
    tag_coeff: float        # coeff to adjust weight of tags
    N_important: int        # number of important landmarks to consider

    preferences: Preferences    # preferences of visit
    location: Tuple[float]      # coordinates around which to find a path

    def __init__(self, preferences: Preferences, coordinates: Tuple[float, float]) -> None:

        with constants.AMENITY_SELECTORS_PATH.open('r') as f:
            self.amenity_selectors = yaml.safe_load(f)

        with constants.LANDMARK_PARAMETERS_PATH.open('r') as f:
            parameters = yaml.safe_load(f)
            self.city_bbox_side = parameters['city_bbox_side']
            self.radius_close_to = parameters['radius_close_to']
            self.church_coeff = parameters['church_coeff']
            self.park_coeff = parameters['park_coeff']
            self.tag_coeff = parameters['tag_coeff']
            self.N_important = parameters['N_important']
        
        self.preferences = preferences
        self.location = coordinates


    def generate_landmarks_list(self) -> Tuple[List[Landmark], List[Landmark]] :
        """
        Generate and prioritize a list of landmarks based on user preferences.

        This method fetches landmarks from various categories (sightseeing, nature, shopping) based on the user's preferences
        and current location. It scores and corrects these landmarks, removes duplicates, and then selects the most important
        landmarks based on a predefined criterion.

        Returns:
            Tuple[List[Landmark], List[Landmark]]:
                - A list of all existing landmarks.
                - A list of the most important landmarks based on the user's preferences.
        """

        L = []

        # List for sightseeing
        if self.preferences.sightseeing.score != 0 :
            L1 = self.fetch_landmarks(self.amenity_selectors['sightseeing'], SIGHTSEEING, coordinates=self.location)
            self.correct_score(L1, self.preferences.sightseeing)
            L += L1
        
        # List for nature
        if self.preferences.nature.score != 0 :
            L2 = self.fetch_landmarks(self.amenity_selectors['nature'], NATURE, coordinates=self.location)
            self.correct_score(L2, self.preferences.nature)
            L += L2
        
        # List for shopping
        if self.preferences.shopping.score != 0 :
            L3 = self.fetch_landmarks(self.amenity_selectors['shopping'], SHOPPING, coordinates=self.location)
            self.correct_score(L3, self.preferences.shopping)
            L += L3

        L = self.remove_duplicates(L)

        return L, take_most_important(L, self.N_important)


    def remove_duplicates(self, landmarks: List[Landmark]) -> List[Landmark] :
        """
        Removes duplicate landmarks based on their names from the given list. Only retains the landmark with highest score

        Parameters:
        landmarks (List[Landmark]): A list of Landmark objects.

        Returns:
        List[Landmark]: A list of unique Landmark objects based on their names.
        """

        L_clean = []
        names = []

        for landmark in landmarks :
            if landmark.name in names: 
                continue  
            else :
                names.append(landmark.name)
                L_clean.append(landmark)
        
        return L_clean
        

    def correct_score(self, landmarks: List[Landmark], preference: Preference) :
        """
        Adjust the attractiveness score of each landmark in the list based on user preferences.

        This method updates the attractiveness of each landmark by scaling it according to the user's preference score.
        The score adjustment is computed using a simple linear transformation based on the preference score.

        Args:
            landmarks (List[Landmark]): A list of landmarks whose scores need to be corrected.
            preference (Preference): The user's preference settings that influence the attractiveness score adjustment.

        Raises:
            TypeError: If the type of any landmark in the list does not match the expected type in the preference.
        """

        if len(landmarks) == 0 :
            return
        
        if landmarks[0].type != preference.type :
            raise TypeError(f"LandmarkType {preference.type} does not match the type of Landmark {landmarks[0].name}")

        for elem in landmarks :
            elem.attractiveness = int(elem.attractiveness*preference.score/5)     # arbitrary computation


    def count_elements_close_to(self, coordinates: Tuple[float, float]) -> int:
        """
        Count the number of OpenStreetMap elements (nodes, ways, relations) within a specified radius of the given location.

        This function constructs a bounding box around the specified coordinates based on the radius. It then queries
        OpenStreetMap data to count the number of elements within that bounding box.

        Args:
            coordinates (Tuple[float, float]): The latitude and longitude of the location to search around.

        Returns:
            int: The number of elements (nodes, ways, relations) within the specified radius. Returns 0 if no elements
                are found or if an error occurs during the query.
        """
        
        lat = coordinates[0]
        lon = coordinates[1]

        radius = self.radius_close_to

        alpha = (180*radius) / (6371000*m.pi)
        bbox = {'latLower':lat-alpha,'lonLower':lon-alpha,'latHigher':lat+alpha,'lonHigher': lon+alpha}
            
        # Build the query to find elements within the radius
        radius_query = overpassQueryBuilder(bbox=[bbox['latLower'],bbox['lonLower'],bbox['latHigher'],bbox['lonHigher']],
                                elementType=['node', 'way', 'relation'])

        try : 
            overpass = Overpass()
            radius_result = overpass.query(radius_query)
            N_elem = radius_result.countWays() + radius_result.countRelations()
            #print(f"There are {N_elem} ways/relations within 50m")
            if N_elem is None :
                return 0
            return N_elem
        except :
            return 0


    def create_bbox(self, coordinates: Tuple[float, float]) -> Tuple[float, float, float, float]:
        """
        Create a bounding box around the given coordinates.

        Args:
            coordinates (Tuple[float, float]): The latitude and longitude of the center of the bounding box.

        Returns:
            Tuple[float, float, float, float]: The minimum latitude, minimum longitude, maximum latitude, and maximum longitude
                                                defining the bounding box.
        """
        
        lat = coordinates[0]
        lon = coordinates[1]

        # Half the side length in km (since it's a square bbox)
        half_side_length_km = self.city_bbox_side / 2 / 1000

        # Convert distance to degrees
        lat_diff = half_side_length_km / 111  # 1 degree latitude is approximately 111 km
        lon_diff = half_side_length_km / (111 * m.cos(m.radians(lat)))  # Adjust for longitude based on latitude

        # Calculate bbox
        min_lat = lat - lat_diff
        max_lat = lat + lat_diff
        min_lon = lon - lon_diff
        max_lon = lon + lon_diff

        return min_lat, min_lon, max_lat, max_lon


    def fetch_landmarks(self, list_amenity: list, landmarktype: str, coordinates: Tuple[float, float]) -> List[Landmark] :
        """
        Fetches landmarks of a specified type from OpenStreetMap (OSM) within a bounding box centered on given coordinates.

        Args:
            list_amenity (list): A list of OSM amenity queries to be used for fetching landmarks. 
                                These queries are typically used to filter results (e.g., [''amenity'='place_of_worship']).
            landmarktype (str): The type of the landmark (e.g., 'sightseeing', 'nature', 'shopping').
            coordinates (Tuple[float, float]): The central coordinates (latitude, longitude) for the bounding box.

        Returns:
            List[Landmark]: A list of Landmark objects that were fetched and filtered based on the provided criteria.

        Notes:
            - The bounding box is created around the given coordinates with a side length defined by `self.city_bbox_side`.
            - Landmarks are fetched using Overpass API queries.
            - Landmarks are filtered based on various conditions including tags and type.
            - Scores are assigned to landmarks based on their attributes and surrounding elements.
        """
        
        # Create bbox around start location
        bbox = self.create_bbox(coordinates)

        # Initialize some variables
        N = 0
        L = []
        overpass = Overpass()

        for amenity in list_amenity :
            query = overpassQueryBuilder(bbox=bbox, elementType=['way', 'relation'], selector=amenity, includeCenter=True, out='body')
            result = overpass.query(query)
            N += result.countElements()

            for elem in result.elements():

                name = elem.tag('name')                             # Add name
                location = (elem.centerLat(), elem.centerLon())     # Add coordinates (lat, lon)

                # skip if unprecise location
                if name is None or location[0] is None:
                    continue

                # skip if unused
                if 'disused:leisure' in elem.tags().keys():
                    continue
                
                # skip if part of another building
                if 'building:part' in elem.tags().keys() and elem.tag('building:part') == 'yes':
                    continue
            
                else :
                    osm_type = elem.type()              # Add type : 'way' or 'relation'
                    osm_id = elem.id()                  # Add OSM id 
                    elem_type = landmarktype            # Add the landmark type as 'sightseeing, 
                    n_tags = len(elem.tags().keys())    # Add number of tags

                    # remove specific tags
                    skip = False
                    for tag in elem.tags().keys() :
                        if "pay" in tag :
                            n_tags -= 1             # discard payment options for tags

                        if "disused" in tag :
                            skip = True             # skip disused amenities
                            break

                        if "wikipedia" in tag :
                            n_tags += 3             # wikipedia entries count more

                        if tag == "wikidata" :
                            Q = elem.tag('wikidata')
                            site = Site("wikidata", "wikidata")
                            item = ItemPage(site, Q)
                            item.get()
                            n_languages = len(item.labels)
                            n_tags += n_languages/10

                        if elem_type != "nature" :
                            if "leisure" in tag and elem.tag('leisure') == "park":
                                elem_type = "nature"

                        if amenity not in ["'shop'='department_store'", "'shop'='mall'"] :
                            if "shop" in tag :
                                skip = True
                                break

                            if tag == "building" and elem.tag('building') in ['retail', 'supermarket', 'parking']:
                                skip = True
                                break

                    if skip:
                        continue

                    # Add score of given landmark based on the number of surrounding elements. Penalty for churches as there are A LOT
                    if amenity == "'amenity'='place_of_worship'" :
                        #score = int((count_elements_close_to(location, radius) + (n_tags*tag_coeff) )*church_coeff)  
                        score = int((self.count_elements_close_to(location) + ((n_tags**1.2)*self.tag_coeff) )*self.church_coeff)  
                    elif amenity == "'leisure'='park'" :
                        score = int((self.count_elements_close_to(location) + ((n_tags**1.2)*self.tag_coeff) )*self.park_coeff)  
                    else :
                        score = int(self.count_elements_close_to(location) + ((n_tags**1.2)*self.tag_coeff))

                    if score is not None :

                        # Generate the landmark and append it to the list
                        #print(f"There are {n_tags} tags on this Landmark. Total score : {score}\n")
                        landmark = Landmark(name=name, type=elem_type, location=location, osm_type=osm_type, osm_id=osm_id, attractiveness=score, must_do=False, n_tags=int(n_tags))
                        L.append(landmark)

        return L


